import copy
from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.common.parameters import Parameters
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.eval.resource_eval import (
    ResourceEval,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.eval.resource_eval_factory import (
    resource_eval_for,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.eval.resource_output_transformer.transformer import (
    ResourceOutputTransformer,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.eval.resource_output_transformer.transformer_factory import (
    resource_output_transformer_for,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.reader_config_decl import (
    ReaderConfig,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    Resource,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class ItemReader(EvalComponent):
    resource_eval: Final[ResourceEval]
    parameters: Final[Optional[Parameters]]
    reader_config: Final[Optional[ReaderConfig]]
    resource_output_transformer: Optional[ResourceOutputTransformer]

    def __init__(
        self,
        resource: Resource,
        parameters: Optional[Parameters],
        reader_config: Optional[ReaderConfig],
    ):
        self.resource_eval = resource_eval_for(resource=resource)
        self.parameters = parameters
        self.reader_config = reader_config

        self.resource_output_transformer = None
        if self.reader_config:
            self.resource_output_transformer = resource_output_transformer_for(
                input_type=self.reader_config.input_type
            )

    @property
    def resource(self):
        return self.resource_eval.resource

    def __str__(self):
        class_dict = copy.deepcopy(self.__dict__)
        del class_dict["resource_eval"]
        class_dict["resource"] = self.resource
        return f"({self.__class__.__name__}| {class_dict})"

    def _eval_body(self, env: Environment) -> None:
        resource_config = None
        if self.reader_config:
            self.reader_config.eval(env=env)
            resource_config = env.stack.pop()
        if self.parameters:
            self.parameters.eval(env=env)
        else:
            env.stack.append(dict())

        self.resource_eval.eval_resource(env=env)

        if self.reader_config:
            env.stack.append(resource_config)
            self.resource_output_transformer.eval(env=env)



import base64
import hashlib
import hmac
import re
import string
from typing import Any, Dict, List, Optional

from moto.moto_api._internal import mock_random as random

FORMATS = {
    "email": r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b",
    "phone_number": r"\+\d{,15}",
}


PAGINATION_MODEL = {
    "list_user_pools": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 60,
        "unique_attribute": "arn",
    },
    "list_user_pool_clients": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 60,
        "unique_attribute": "id",
    },
    "list_identity_providers": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 60,
        "unique_attribute": "name",
    },
    "list_users": {
        "input_token": "pagination_token",
        "limit_key": "limit",
        "limit_default": 60,
        "unique_attribute": "id",
    },
    "list_groups": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 60,
        "unique_attribute": "group_name",
    },
    "admin_list_groups_for_user": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 60,
        "unique_attribute": "group_name",
    },
    "list_users_in_group": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 60,
        "unique_attribute": "id",
    },
    "list_resource_servers": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 60,
        "unique_attribute": "identifier",
    },
}


def create_id() -> str:
    size = 26
    chars = list(range(10)) + list(string.ascii_lowercase)
    return "".join(str(random.choice(chars)) for x in range(size))


def check_secret_hash(
    app_client_secret: str,
    app_client_id: str,
    username: str,
    secret_hash: Optional[str],
) -> bool:
    key = bytes(str(app_client_secret).encode("latin-1"))
    msg = bytes(str(username + app_client_id).encode("latin-1"))
    new_digest = hmac.new(key, msg, hashlib.sha256).digest()
    SECRET_HASH = base64.b64encode(new_digest).decode()
    return SECRET_HASH == secret_hash


def validate_username_format(username: str, _format: str = "email") -> bool:
    # if the value of the `_format` param other than `email` or `phone_number`,
    # the default value for the regex will match nothing and the
    # method will return None
    return re.fullmatch(FORMATS.get(_format, r"a^"), username) is not None


def flatten_attrs(attrs: List[Dict[str, Any]]) -> Dict[str, Any]:
    return {attr["Name"]: attr["Value"] for attr in attrs}


def expand_attrs(attrs: Dict[str, Any]) -> List[Dict[str, Any]]:
    return [{"Name": k, "Value": v} for k, v in attrs.items()]


ID_HASH_STRATEGY = "HASH"


def generate_id(strategy: Optional[str], *args: Any) -> str:
    if strategy == ID_HASH_STRATEGY:
        return _generate_id_hash(args)
    else:
        return _generate_id_uuid()


def _generate_id_uuid() -> str:
    return random.uuid4().hex


def _generate_id_hash(args: Any) -> str:
    hasher = hashlib.sha256()

    for arg in args:
        hasher.update(str(arg).encode())

    return hasher.hexdigest()



import json
from uuid import uuid4

import boto3

from moto import mock_aws


@mock_aws
def test_create_basic_stack():
    # Create inner template
    cf = boto3.client("cloudformation", "us-east-1")
    bucket_created_by_cf = str(uuid4())
    template = get_inner_template(bucket_created_by_cf)
    # Upload inner template to S3
    s3 = boto3.client("s3", "us-east-1")
    cf_storage_bucket = str(uuid4())
    s3.create_bucket(Bucket=cf_storage_bucket)
    s3.put_object(Bucket=cf_storage_bucket, Key="stack.json", Body=json.dumps(template))

    # Create template that includes the inner template
    stack_name = "a" + str(uuid4())[0:6]
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Resources": {
            "NestedStack": {
                "Type": "AWS::CloudFormation::Stack",
                "Properties": {
                    "TemplateURL": f"https://s3.amazonaws.com/{cf_storage_bucket}/stack.json",
                },
            },
        },
    }
    cf.create_stack(StackName=stack_name, TemplateBody=str(template))

    # Verify the inner S3 bucket has been created
    bucket_names = sorted([b["Name"] for b in s3.list_buckets()["Buckets"]])
    assert bucket_names == sorted([cf_storage_bucket, bucket_created_by_cf])

    # Verify both stacks are created
    stacks = cf.list_stacks()["StackSummaries"]
    assert len(stacks) == 2


@mock_aws
def test_create_stack_with_params():
    # Create inner template
    cf = boto3.client("cloudformation", "us-east-1")
    bucket_created_by_cf = str(uuid4())
    inner_template = json.dumps(get_inner_template_with_params())

    # Upload inner template to S3
    s3 = boto3.client("s3", "us-east-1")
    cf_storage_bucket = str(uuid4())
    s3.create_bucket(Bucket=cf_storage_bucket)
    s3.put_object(Bucket=cf_storage_bucket, Key="stack.json", Body=inner_template)

    # Create template that includes the inner template
    stack_name = "a" + str(uuid4())[0:6]
    template = get_outer_template_with_params(cf_storage_bucket, bucket_created_by_cf)
    cf.create_stack(StackName=stack_name, TemplateBody=str(template))

    # Verify the inner S3 bucket has been created
    bucket_names = sorted([b["Name"] for b in s3.list_buckets()["Buckets"]])
    assert bucket_names == sorted([cf_storage_bucket, bucket_created_by_cf])


@mock_aws
def test_update_stack_with_params():
    # Create inner template
    cf = boto3.client("cloudformation", "us-east-1")
    first_bucket = str(uuid4())
    second_bucket = str(uuid4())
    inner_template = json.dumps(get_inner_template_with_params())

    # Upload inner template to S3
    s3 = boto3.client("s3", "us-east-1")
    cf_storage_bucket = str(uuid4())
    s3.create_bucket(Bucket=cf_storage_bucket)
    s3.put_object(Bucket=cf_storage_bucket, Key="stack.json", Body=inner_template)

    # Create template that includes the inner template
    stack_name = "a" + str(uuid4())[0:6]
    template = get_outer_template_with_params(cf_storage_bucket, first_bucket)
    cf.create_stack(StackName=stack_name, TemplateBody=str(template))

    # Verify the inner S3 bucket has been created
    bucket_names = sorted([b["Name"] for b in s3.list_buckets()["Buckets"]])
    assert bucket_names == sorted([cf_storage_bucket, first_bucket])

    # Update stack
    template = get_outer_template_with_params(cf_storage_bucket, second_bucket)
    cf.update_stack(StackName=stack_name, TemplateBody=str(template))

    # Verify the inner S3 bucket has been created
    bucket_names = sorted([b["Name"] for b in s3.list_buckets()["Buckets"]])
    assert bucket_names == sorted([cf_storage_bucket, second_bucket])


@mock_aws
def test_delete_basic_stack():
    # Create inner template
    cf = boto3.client("cloudformation", "us-east-1")
    bucket_created_by_cf = str(uuid4())
    template = get_inner_template(bucket_created_by_cf)

    # Upload inner template to S3
    s3 = boto3.client("s3", "us-east-1")
    cf_storage_bucket = str(uuid4())
    s3.create_bucket(Bucket=cf_storage_bucket)
    s3.put_object(Bucket=cf_storage_bucket, Key="stack.json", Body=json.dumps(template))

    # Create template that includes the inner template
    stack_name = "a" + str(uuid4())[0:6]
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Resources": {
            "NestedStack": {
                "Type": "AWS::CloudFormation::Stack",
                "Properties": {
                    "TemplateURL": f"https://s3.amazonaws.com/{cf_storage_bucket}/stack.json",
                },
            },
        },
    }
    cf.create_stack(StackName=stack_name, TemplateBody=str(template))
    cf.delete_stack(StackName=stack_name)

    # Verify the stack-controlled S3 bucket has been deleted
    bucket_names = sorted([b["Name"] for b in s3.list_buckets()["Buckets"]])
    assert bucket_names == [cf_storage_bucket]

    # Verify both stacks are deleted
    stacks = cf.list_stacks()["StackSummaries"]
    assert len(stacks) == 2
    for stack in stacks:
        assert stack["StackStatus"] == "DELETE_COMPLETE"


def get_inner_template(bucket_created_by_cf):
    return {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Resources": {
            "bcbcf": {
                "Type": "AWS::S3::Bucket",
                "Properties": {"BucketName": bucket_created_by_cf},
            }
        },
        "Outputs": {"Bucket": {"Value": {"Ref": "bcbcf"}}},
    }


def get_inner_template_with_params():
    return {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Parameters": {
            "BName": {"Description": "bucket name", "Type": "String"},
        },
        "Resources": {
            "bcbcf": {
                "Type": "AWS::S3::Bucket",
                "Properties": {"BucketName": {"Ref": "BName"}},
            }
        },
    }


def get_outer_template_with_params(cf_storage_bucket, first_bucket):
    return {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Resources": {
            "NestedStack": {
                "Type": "AWS::CloudFormation::Stack",
                "Properties": {
                    "TemplateURL": f"https://s3.amazonaws.com/{cf_storage_bucket}/stack.json",
                    "Parameters": {"BName": first_bucket},
                },
            },
        },
    }



import threading


class CountDownLatch:
    # TODO: add timeout support.
    def __init__(self, num: int):
        self._num: int = num
        self.lock = threading.Condition()

    def count_down(self) -> None:
        self.lock.acquire()
        self._num -= 1
        if self._num <= 0:
            self.lock.notify_all()
        self.lock.release()

    def wait(self) -> None:
        self.lock.acquire()
        while self._num > 0:
            self.lock.wait()
        self.lock.release()



"""Handles incoming appmesh requests, invokes methods, returns responses."""

import json

from moto.appmesh.utils.spec_parsing import (
    build_route_spec,
    build_virtual_node_spec,
    port_mappings_from_router_spec,
)
from moto.core.responses import BaseResponse

from .models import AppMeshBackend, appmesh_backends


class AppMeshResponse(BaseResponse):
    """Handler for AppMesh requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="appmesh")

    @property
    def appmesh_backend(self) -> AppMeshBackend:
        """Return backend instance specific for this region."""
        return appmesh_backends[self.current_account][self.region]

    def create_mesh(self) -> str:
        params = json.loads(self.body)
        client_token = params.get("clientToken")
        mesh_name = params.get("meshName")
        spec = params.get("spec") or {}
        egress_filter_type = (spec.get("egressFilter") or {}).get("type")
        ip_preference = (spec.get("serviceDiscovery") or {}).get("ipPreference")
        tags = params.get("tags")
        mesh = self.appmesh_backend.create_mesh(
            client_token=client_token,
            mesh_name=mesh_name,
            egress_filter_type=egress_filter_type,
            ip_preference=ip_preference,
            tags=tags,
        )
        return json.dumps(mesh.to_dict())

    def update_mesh(self) -> str:
        params = json.loads(self.body)
        client_token = params.get("clientToken")
        mesh_name = self._get_param("meshName")
        spec = params.get("spec") or {}
        egress_filter_type = (spec.get("egressFilter") or {}).get("type")
        ip_preference = (spec.get("serviceDiscovery") or {}).get("ipPreference")
        mesh = self.appmesh_backend.update_mesh(
            client_token=client_token,
            mesh_name=mesh_name,
            egress_filter_type=egress_filter_type,
            ip_preference=ip_preference,
        )
        return json.dumps(mesh.to_dict())

    def describe_mesh(self) -> str:
        mesh_name = self._get_param(param_name="meshName", if_none="")
        mesh_owner = self._get_param("meshOwner")
        mesh = self.appmesh_backend.describe_mesh(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
        )
        return json.dumps(mesh.to_dict())

    def delete_mesh(self) -> str:
        mesh_name = self._get_param("meshName")
        mesh = self.appmesh_backend.delete_mesh(mesh_name=mesh_name)
        return json.dumps(mesh.to_dict())

    def list_meshes(self) -> str:
        params = self._get_params()
        limit = params.get("limit")
        next_token = params.get("nextToken")
        meshes, next_token = self.appmesh_backend.list_meshes(
            limit=limit,
            next_token=next_token,
        )
        return json.dumps(dict(meshes=meshes, nextToken=next_token))

    def list_tags_for_resource(self) -> str:
        params = self._get_params()
        limit = params.get("limit")
        next_token = params.get("nextToken")
        resource_arn = params.get("resourceArn")
        tags, next_token = self.appmesh_backend.list_tags_for_resource(
            limit=limit,
            next_token=next_token,
            resource_arn=resource_arn,
        )
        return json.dumps(dict(nextToken=next_token, tags=tags))

    def tag_resource(self) -> str:
        params = json.loads(self.body)
        resource_arn = self._get_param("resourceArn")
        tags = params.get("tags")
        self.appmesh_backend.tag_resource(
            resource_arn=resource_arn,
            tags=tags,
        )
        return json.dumps(dict())

    def describe_virtual_router(self) -> str:
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        virtual_router_name = self._get_param("virtualRouterName")
        virtual_router = self.appmesh_backend.describe_virtual_router(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
        )
        return json.dumps(virtual_router.to_dict())

    def create_virtual_router(self) -> str:
        params = json.loads(self.body)
        client_token = params.get("clientToken")
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        port_mappings = port_mappings_from_router_spec(params.get("spec"))
        tags = params.get("tags")
        virtual_router_name = params.get("virtualRouterName")
        virtual_router = self.appmesh_backend.create_virtual_router(
            client_token=client_token,
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            port_mappings=port_mappings,
            tags=tags,
            virtual_router_name=virtual_router_name,
        )
        return json.dumps(virtual_router.to_dict())

    def update_virtual_router(self) -> str:
        params = json.loads(self.body)
        client_token = params.get("clientToken")
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        port_mappings = port_mappings_from_router_spec(params.get("spec"))
        virtual_router_name = self._get_param("virtualRouterName")
        virtual_router = self.appmesh_backend.update_virtual_router(
            client_token=client_token,
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            port_mappings=port_mappings,
            virtual_router_name=virtual_router_name,
        )
        return json.dumps(virtual_router.to_dict())

    def delete_virtual_router(self) -> str:
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        virtual_router_name = self._get_param("virtualRouterName")
        virtual_router = self.appmesh_backend.delete_virtual_router(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
        )
        return json.dumps(virtual_router.to_dict())

    def list_virtual_routers(self) -> str:
        limit = self._get_param("limit")
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        next_token = self._get_param("nextToken")
        virtual_routers, next_token = self.appmesh_backend.list_virtual_routers(
            limit=limit,
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            next_token=next_token,
        )
        return json.dumps(dict(nextToken=next_token, virtualRouters=virtual_routers))

    def create_route(self) -> str:
        params = json.loads(self.body)
        client_token = params.get("clientToken")
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        route_name = self._get_param("routeName")
        tags = params.get("tags")
        virtual_router_name = self._get_param("virtualRouterName")
        spec = build_route_spec(params.get("spec") or {})
        route = self.appmesh_backend.create_route(
            client_token=client_token,
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            route_name=route_name,
            spec=spec,
            tags=tags,
            virtual_router_name=virtual_router_name,
        )
        return json.dumps(route.to_dict())

    def describe_route(self) -> str:
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        route_name = self._get_param("routeName")
        virtual_router_name = self._get_param("virtualRouterName")
        route = self.appmesh_backend.describe_route(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            route_name=route_name,
            virtual_router_name=virtual_router_name,
        )
        return json.dumps(route.to_dict())

    def update_route(self) -> str:
        params = json.loads(self.body)
        client_token = params.get("clientToken")
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        route_name = self._get_param("routeName")
        virtual_router_name = self._get_param("virtualRouterName")
        spec = build_route_spec(params.get("spec") or {})
        route = self.appmesh_backend.update_route(
            client_token=client_token,
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            route_name=route_name,
            spec=spec,
            virtual_router_name=virtual_router_name,
        )
        return json.dumps(route.to_dict())

    def delete_route(self) -> str:
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        route_name = self._get_param("routeName")
        virtual_router_name = self._get_param("virtualRouterName")
        route = self.appmesh_backend.delete_route(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            route_name=route_name,
            virtual_router_name=virtual_router_name,
        )
        return json.dumps(route.to_dict())

    def list_routes(self) -> str:
        limit = self._get_param("limit")
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        next_token = self._get_param("nextToken")
        virtual_router_name = self._get_param("virtualRouterName")
        routes, next_token = self.appmesh_backend.list_routes(
            limit=limit,
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            next_token=next_token,
            virtual_router_name=virtual_router_name,
        )
        return json.dumps(dict(nextToken=next_token, routes=routes))

    def describe_virtual_node(self) -> str:
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        virtual_node_name = self._get_param("virtualNodeName")
        virtual_node = self.appmesh_backend.describe_virtual_node(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_node_name=virtual_node_name,
        )
        return json.dumps(virtual_node.to_dict())

    def create_virtual_node(self) -> str:
        params = json.loads(self.body)
        client_token = params.get("clientToken")
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        spec = build_virtual_node_spec(params.get("spec") or {})
        tags = params.get("tags")
        virtual_node_name = params.get("virtualNodeName")
        virtual_node = self.appmesh_backend.create_virtual_node(
            client_token=client_token,
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            spec=spec,
            tags=tags,
            virtual_node_name=virtual_node_name,
        )
        return json.dumps(virtual_node.to_dict())

    def update_virtual_node(self) -> str:
        params = json.loads(self.body)
        client_token = params.get("clientToken")
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        spec = build_virtual_node_spec(params.get("spec") or {})
        virtual_node_name = self._get_param("virtualNodeName")
        virtual_node = self.appmesh_backend.update_virtual_node(
            client_token=client_token,
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            spec=spec,
            virtual_node_name=virtual_node_name,
        )
        return json.dumps(virtual_node.to_dict())

    def delete_virtual_node(self) -> str:
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        virtual_node_name = self._get_param("virtualNodeName")
        virtual_node = self.appmesh_backend.delete_virtual_node(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_node_name=virtual_node_name,
        )
        return json.dumps(virtual_node.to_dict())

    def list_virtual_nodes(self) -> str:
        limit = self._get_param("limit")
        mesh_name = self._get_param("meshName")
        mesh_owner = self._get_param("meshOwner")
        next_token = self._get_param("nextToken")
        virtual_nodes, next_token = self.appmesh_backend.list_virtual_nodes(
            limit=limit,
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            next_token=next_token,
        )
        return json.dumps(dict(nextToken=next_token, virtualNodes=virtual_nodes))



import boto3

from moto import mock_aws


@mock_aws
def test_describe():
    client = boto3.client("neptune", "us-east-2")
    assert client.describe_global_clusters()["GlobalClusters"] == []


@mock_aws
def test_create_global_cluster():
    client = boto3.client("neptune", "us-east-1")
    resp = client.create_global_cluster(
        GlobalClusterIdentifier="g-id", Engine="neptune"
    )["GlobalCluster"]
    assert resp["GlobalClusterIdentifier"] == "g-id"
    assert "GlobalClusterResourceId" in resp
    assert "GlobalClusterArn" in resp
    assert resp["Engine"] == "neptune"
    assert "EngineVersion" in resp
    assert resp["StorageEncrypted"] is False
    assert resp["DeletionProtection"] is False

    assert len(client.describe_global_clusters()["GlobalClusters"]) == 1

    # As a global cluster, verify it can be retrieved everywhere
    europe_client = boto3.client("neptune", "eu-north-1")
    assert len(europe_client.describe_global_clusters()["GlobalClusters"]) == 1


@mock_aws
def test_create_global_cluster_with_additional_params():
    client = boto3.client("neptune", "us-east-1")
    resp = client.create_global_cluster(
        GlobalClusterIdentifier="g-id",
        Engine="neptune",
        EngineVersion="1.0",
        DeletionProtection=True,
        StorageEncrypted=True,
    )["GlobalCluster"]
    assert resp["Engine"] == "neptune"
    assert resp["EngineVersion"] == "1.0"
    assert resp["StorageEncrypted"] is True
    assert resp["DeletionProtection"] is True


@mock_aws
def test_delete_global_cluster():
    client = boto3.client("neptune", "us-east-2")
    client.create_global_cluster(GlobalClusterIdentifier="g-id2", Engine="neptune")

    client.delete_global_cluster(GlobalClusterIdentifier="g-id2")

    assert client.describe_global_clusters()["GlobalClusters"] == []



from moto.ec2.exceptions import InvalidParameterValueErrorUnknownAttribute
from moto.ec2.utils import add_tag_specification, get_attribute_value

from ._base_response import EC2BaseResponse


class ElasticNetworkInterfaces(EC2BaseResponse):
    def create_network_interface(self) -> str:
        subnet_id = self._get_param("SubnetId")
        private_ip_address = self._get_param("PrivateIpAddress")
        private_ip_addresses = self._get_multi_param("PrivateIpAddresses")
        ipv6_addresses = self._get_multi_param("Ipv6Addresses")
        ipv6_address_count = self._get_int_param("Ipv6AddressCount", 0)
        secondary_ips_count = self._get_param("SecondaryPrivateIpAddressCount")
        groups = self._get_multi_param("SecurityGroupId")
        subnet = self.ec2_backend.get_subnet(subnet_id)
        description = self._get_param("Description")
        tags = add_tag_specification(self._get_multi_param("TagSpecification"))

        self.error_on_dryrun()

        eni = self.ec2_backend.create_network_interface(
            subnet,
            private_ip_address,
            private_ip_addresses,
            groups,
            description,
            tags,
            secondary_ips_count=secondary_ips_count,
            ipv6_addresses=ipv6_addresses,
            ipv6_address_count=ipv6_address_count,
        )
        template = self.response_template(CREATE_NETWORK_INTERFACE_RESPONSE)
        return template.render(eni=eni)

    def delete_network_interface(self) -> str:
        eni_id = self._get_param("NetworkInterfaceId")

        self.error_on_dryrun()

        self.ec2_backend.delete_network_interface(eni_id)
        template = self.response_template(DELETE_NETWORK_INTERFACE_RESPONSE)
        return template.render()

    def describe_network_interface_attribute(self) -> str:
        eni_id = self._get_param("NetworkInterfaceId")
        attribute = self._get_param("Attribute")

        self.error_on_dryrun()

        eni = self.ec2_backend.get_all_network_interfaces([eni_id])[0]

        if attribute == "description":
            template = self.response_template(
                DESCRIBE_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE_DESCRIPTION
            )
        elif attribute == "groupSet":
            template = self.response_template(
                DESCRIBE_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE_GROUPSET
            )
        elif attribute == "sourceDestCheck":
            template = self.response_template(
                DESCRIBE_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE_SOURCEDESTCHECK
            )
        elif attribute == "attachment":
            template = self.response_template(
                DESCRIBE_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE_ATTACHMENT
            )
        else:
            raise InvalidParameterValueErrorUnknownAttribute(attribute)
        return template.render(eni=eni)

    def describe_network_interfaces(self) -> str:
        eni_ids = self._get_multi_param("NetworkInterfaceId")
        filters = self._filters_from_querystring()

        self.error_on_dryrun()

        enis = self.ec2_backend.get_all_network_interfaces(eni_ids, filters)
        template = self.response_template(DESCRIBE_NETWORK_INTERFACES_RESPONSE)
        return template.render(enis=enis)

    def attach_network_interface(self) -> str:
        eni_id = self._get_param("NetworkInterfaceId")
        instance_id = self._get_param("InstanceId")
        device_index = self._get_param("DeviceIndex")

        self.error_on_dryrun()

        attachment_id = self.ec2_backend.attach_network_interface(
            eni_id, instance_id, device_index
        )
        template = self.response_template(ATTACH_NETWORK_INTERFACE_RESPONSE)
        return template.render(attachment_id=attachment_id)

    def detach_network_interface(self) -> str:
        attachment_id = self._get_param("AttachmentId")

        self.error_on_dryrun()

        self.ec2_backend.detach_network_interface(attachment_id)
        template = self.response_template(DETACH_NETWORK_INTERFACE_RESPONSE)
        return template.render()

    def modify_network_interface_attribute(self) -> str:
        eni_id = self._get_param("NetworkInterfaceId")
        group_ids = self._get_multi_param("SecurityGroupId")
        source_dest_check = get_attribute_value("SourceDestCheck", self.querystring)
        description = get_attribute_value("Description", self.querystring)

        self.error_on_dryrun()

        self.ec2_backend.modify_network_interface_attribute(
            eni_id, group_ids, source_dest_check, description
        )
        return MODIFY_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE

    def reset_network_interface_attribute(self) -> str:
        self.error_on_dryrun()

        raise NotImplementedError(
            "ElasticNetworkInterfaces(AmazonVPC).reset_network_interface_attribute is not yet implemented"
        )

    def assign_private_ip_addresses(self) -> str:
        eni_id = self._get_param("NetworkInterfaceId")
        secondary_ips_count = self._get_int_param("SecondaryPrivateIpAddressCount", 0)
        private_ip_addresses = self._get_multi_param("PrivateIpAddress")
        eni = self.ec2_backend.assign_private_ip_addresses(
            eni_id, private_ip_addresses, secondary_ips_count
        )
        template = self.response_template(ASSIGN_PRIVATE_IP_ADDRESSES)
        return template.render(eni=eni)

    def unassign_private_ip_addresses(self) -> str:
        eni_id = self._get_param("NetworkInterfaceId")
        private_ip_address = self._get_multi_param("PrivateIpAddress")
        eni = self.ec2_backend.unassign_private_ip_addresses(eni_id, private_ip_address)
        template = self.response_template(UNASSIGN_PRIVATE_IP_ADDRESSES)
        return template.render(eni=eni)

    def assign_ipv6_addresses(self) -> str:
        eni_id = self._get_param("NetworkInterfaceId")
        ipv6_count = self._get_int_param("Ipv6AddressCount", 0)
        ipv6_addresses = self._get_multi_param("Ipv6Addresses")
        eni = self.ec2_backend.assign_ipv6_addresses(eni_id, ipv6_addresses, ipv6_count)
        template = self.response_template(ASSIGN_IPV6_ADDRESSES)
        return template.render(eni=eni)

    def unassign_ipv6_addresses(self) -> str:
        eni_id = self._get_param("NetworkInterfaceId")
        ips = self._get_multi_param("Ipv6Addresses")
        eni = self.ec2_backend.unassign_ipv6_addresses(eni_id, ips)
        template = self.response_template(UNASSIGN_IPV6_ADDRESSES)
        return template.render(eni=eni, unassigned_ips=ips)


ASSIGN_PRIVATE_IP_ADDRESSES = """<AssignPrivateIpAddressesResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
  <requestId>3fb591ba-558c-48f8-ae6b-c2f9d6d06425</requestId>
  <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
  <assignedPrivateIpAddressesSet>
    {% for address in eni.private_ip_addresses %}
    <item>
        <privateIpAddress>{{ address.PrivateIpAddress }}</privateIpAddress>
    </item>
    {% endfor %}
  </assignedPrivateIpAddressesSet>
  <return>true</return>
</AssignPrivateIpAddressesResponse>"""


UNASSIGN_PRIVATE_IP_ADDRESSES = """<UnAssignPrivateIpAddressesResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
  <requestId>3fb591ba-558c-48f8-ae6b-c2f9d6d06425</requestId>
  <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
  <assignedPrivateIpAddressesSet>
    {% for address in eni.private_ip_addresses %}
    <item>
        <privateIpAddress>{{ address.PrivateIpAddress }}</privateIpAddress>
    </item>
    {% endfor %}
  </assignedPrivateIpAddressesSet>
  <return>true</return>
</UnAssignPrivateIpAddressesResponse>"""


ASSIGN_IPV6_ADDRESSES = """<AssignIpv6AddressesResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
    <requestId>c36d17eb-a0ba-4d38-8727-example</requestId>
    <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
    <assignedIpv6Addresses>
        {% for address in eni.ipv6_addresses %}
        <item>{{address}}</item>
        {% endfor %}
    </assignedIpv6Addresses>
</AssignIpv6AddressesResponse>
"""

UNASSIGN_IPV6_ADDRESSES = """<UnassignIpv6AddressesResponse xmlns="http://ec2.amazonaws.com/doc/2016-11-15/">
    <requestId>94d446d7-fc8e-4918-94f9-example</requestId>
    <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
    <unassignedIpv6Addresses>
        {% for address in unassigned_ips %}
        <item>{{address}}</item>
        {% endfor %}
    </unassignedIpv6Addresses>
</UnassignIpv6AddressesResponse>"""


CREATE_NETWORK_INTERFACE_RESPONSE = """
<CreateNetworkInterfaceResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
    <requestId>2c6021ec-d705-445a-9780-420d0c7ab793</requestId>
    <networkInterface>
        <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
        <subnetId>{{ eni.subnet.id }}</subnetId>
        <vpcId>{{ eni.subnet.vpc_id }}</vpcId>
        <availabilityZone>{{ eni.subnet.availability_zone }}</availabilityZone>
        {% if eni.description %}
        <description>{{ eni.description }}</description>
        {% endif %}
        <ownerId>{{ eni.owner_id }}</ownerId>
        <requesterId>AIDARCSPW2WNREUEN7XFM</requesterId>
        <requesterManaged>False</requesterManaged>
        <status>{{ eni.status }}</status>
        <macAddress>{{ eni.mac_address }}</macAddress>
        {% if eni.private_ip_address %}
          <privateIpAddress>{{ eni.private_ip_address }}</privateIpAddress>
        {% endif %}
        {% if eni.private_dns_name %}
        <privateDnsName>{{ eni.private_dns_name }}</privateDnsName>
        {% endif %}
        <sourceDestCheck>{{ "true" if eni.source_dest_check == True else "false" }}</sourceDestCheck>
        <groupSet>
        {% for group in eni.group_set %}
            <item>
                <groupId>{{ group.id }}</groupId>
                <groupName>{{ group.name }}</groupName>
             </item>
         {% endfor %}
         </groupSet>
        {% if eni.association %}
        <association>
            <publicIp>{{ eni.public_ip }}</publicIp>
            <ipOwnerId>{{ eni.owner_id }}</ipOwnerId>
            <allocationId>{{ eni.association.allocationId }}</allocationId>
            <associationId>{{ eni.association.associationId }}</associationId>
            <natEnabled>true</natEnabled>
        </association>
        {% endif %}
        <tagSet>
          {% for tag in eni.get_tags() %}
              <item>
                  <key>{{ tag.key }}</key>
                  <value>{{ tag.value }}</value>
              </item>
          {% endfor %}
        </tagSet>
        <privateIpAddressesSet>
          {% for address in eni.private_ip_addresses %}
          <item>
            <privateIpAddress>{{ address.PrivateIpAddress }}</privateIpAddress>
            {% if address.privateDnsName %}
            <privateDnsName>{{ address.PrivateDnsName }}</privateDnsName>
            {% endif %}
            <primary>{{ "true" if address.Primary == True else "false" }}</primary>
          </item>
          {% endfor %}
        </privateIpAddressesSet>
        <ipv6AddressesSet>
        {% for address in eni.ipv6_addresses %}
            <item>
                <ipv6Address>{{address}}</ipv6Address>
            </item>
        {% endfor %}
        </ipv6AddressesSet>
        <interfaceType>{{ eni.interface_type }}</interfaceType>
    </networkInterface>
</CreateNetworkInterfaceResponse>
"""

DESCRIBE_NETWORK_INTERFACES_RESPONSE = """<DescribeNetworkInterfacesResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
    <requestId>ddb0aaf1-8b65-4f0a-94fa-654b18b8a204</requestId>
    <networkInterfaceSet>
    {% for eni in enis %}
        <item>
            <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
            <subnetId>{{ eni.subnet.id }}</subnetId>
            <vpcId>{{ eni.subnet.vpc_id }}</vpcId>
            <availabilityZone>{{ eni.subnet.availability_zone }}</availabilityZone>
            {% if eni.description %}
            <description>{{ eni.description }}</description>
            {% endif %}
            <ownerId>{{ eni.owner_id }}</ownerId>
            <requesterId>AIDARCSPW2WNREUEN7XFM</requesterId>
            <requesterManaged>False</requesterManaged>
            <status>{{ eni.status }}</status>
            <macAddress>{{ eni.mac_address }}</macAddress>
            {% if eni.private_ip_address %}
            <privateIpAddress>{{ eni.private_ip_address }}</privateIpAddress>
            {% endif %}
            {% if eni.private_dns_name %}
            <privateDnsName>{{ eni.private_dns_name }}</privateDnsName>
            {% endif %}
            <sourceDestCheck>{{ "true" if eni.source_dest_check == True else "false" }}</sourceDestCheck>
            <groupSet>
            {% for group in eni.group_set %}
                <item>
                    <groupId>{{ group.id }}</groupId>
                    <groupName>{{ group.name }}</groupName>
                </item>
            {% endfor %}
            </groupSet>
            {% if eni.association %}
            <association>
                <publicIp>{{ eni.public_ip }}</publicIp>
                <ipOwnerId>{{ eni.owner_id }}</ipOwnerId>
                <allocationId>{{ eni.association.allocationId }}</allocationId>
                <associationId>{{ eni.association.associationId }}</associationId>
                <natEnabled>true</natEnabled>
            </association>
            {% endif %}
            {% if eni.attachment_id %}
            <attachment>
                <attachTime>{{ eni.attach_time }}</attachTime>
                <attachmentId>{{ eni.attachment_id }}</attachmentId>
                <deleteOnTermination>{{ 'true' if eni.delete_on_termination else 'false' }}</deleteOnTermination>
                <deviceIndex>{{ eni.device_index }}</deviceIndex>
                <networkCardIndex>0</networkCardIndex>
                <instanceId>{{ eni.instance.id }}</instanceId>
                <instanceOwnerId>{{ eni.instance.owner_id }}</instanceOwnerId>
                <status>attached</status>
            </attachment>
            {% endif %}
            <tagSet>
            {% for tag in eni.get_tags() %}
                <item>
                    <key>{{ tag.key }}</key>
                    <value>{{ tag.value }}</value>
                </item>
            {% endfor %}
            </tagSet>
            <privateIpAddressesSet>
            {% for address in eni.private_ip_addresses %}
            <item>
                <privateIpAddress>{{ address.PrivateIpAddress }}</privateIpAddress>
                {% if address.privateDnsName %}
                <privateDnsName>{{ address.PrivateDnsName }}</privateDnsName>
                {% endif %}
                <primary>{{ "true" if address.Primary == True else "false" }}</primary>
            </item>
            {% endfor %}
            </privateIpAddressesSet>
            <ipv6AddressesSet>
                {% for address in eni.ipv6_addresses %}
                <item>
                    <ipv6Address>{{address}}</ipv6Address>
                </item>
                {% endfor %}
            </ipv6AddressesSet>
            <interfaceType>{{ eni.interface_type }}</interfaceType>
        </item>
    {% endfor %}
    </networkInterfaceSet>
</DescribeNetworkInterfacesResponse>"""

ATTACH_NETWORK_INTERFACE_RESPONSE = """<AttachNetworkInterfaceResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <attachmentId>{{ attachment_id }}</attachmentId>
</AttachNetworkInterfaceResponse>"""

DETACH_NETWORK_INTERFACE_RESPONSE = """<DetachNetworkInterfaceResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</DetachNetworkInterfaceResponse>"""

MODIFY_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE = """<ModifyNetworkInterfaceAttributeResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
  <requestId>59dbff89-35bd-4eac-99ed-be587EXAMPLE</requestId>
  <return>true</return>
</ModifyNetworkInterfaceAttributeResponse>"""

DELETE_NETWORK_INTERFACE_RESPONSE = """
<DeleteNetworkInterfaceResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
    <requestId>34b5b3b4-d0c5-49b9-b5e2-a468ef6adcd8</requestId>
    <return>true</return>
</DeleteNetworkInterfaceResponse>"""

DESCRIBE_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE_DESCRIPTION = """
<DescribeNetworkInterfaceAttributeResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
    <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
    <description>
        <value>{{ eni.description }}</value>
    </description>
</DescribeNetworkInterfaceAttributeResponse>"""

DESCRIBE_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE_GROUPSET = """
<DescribeNetworkInterfaceAttributeResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
    <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
    <groupSet>
        {% for group in eni.group_set %}
        <item>
            <groupId>{{ group.id }}</groupId>
            <groupName>{{ group.name }}</groupName>
        </item>
        {% endfor %}
    </groupSet>
</DescribeNetworkInterfaceAttributeResponse>"""

DESCRIBE_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE_SOURCEDESTCHECK = """
<DescribeNetworkInterfaceAttributeResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
    <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
    <sourceDestCheck>
        <value>{{ "true" if eni.source_dest_check == True else "false" }}</value>
    </sourceDestCheck>
</DescribeNetworkInterfaceAttributeResponse>"""

DESCRIBE_NETWORK_INTERFACE_ATTRIBUTE_RESPONSE_ATTACHMENT = """
<DescribeNetworkInterfaceAttributeResponse xmlns="http://ec2.amazonaws.com/doc/2013-10-15/">
    <networkInterfaceId>{{ eni.id }}</networkInterfaceId>
    {% if eni.attachment_id %}
    <attachment>
        <attachTime>{{ eni.attach_time }}</attachTime>
        <attachmentId>{{ eni.attachment_id }}</attachmentId>
        <deleteOnTermination>{{ 'true' if eni.delete_on_termination else 'false' }}</deleteOnTermination>
        <deviceIndex>{{ eni.device_index }}</deviceIndex>
        <networkCardIndex>0</networkCardIndex>
        <instanceId>{{ eni.instance.id }}</instanceId>
        <instanceOwnerId>{{ eni.instance.owner_id }}</instanceOwnerId>
        <status>attached</status>
    </attachment>
    {% endif %}
</DescribeNetworkInterfaceAttributeResponse>"""



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_service_minimal():
    client = boto3.client("servicediscovery", region_name="ap-southeast-1")
    operation_id = client.create_http_namespace(Name="mynamespace")["OperationId"]
    namespace_id = client.get_operation(OperationId=operation_id)["Operation"][
        "Targets"
    ]["NAMESPACE"]

    resp = client.create_service(Name="my service", NamespaceId=namespace_id)

    assert "Service" in resp
    assert "Id" in resp["Service"]
    assert "Arn" in resp["Service"]
    assert resp["Service"]["Name"] == "my service"
    assert resp["Service"]["NamespaceId"] == namespace_id
    assert "CreateDate" in resp["Service"]


@mock_aws
def test_create_service():
    client = boto3.client("servicediscovery", region_name="ap-southeast-1")
    operation_id = client.create_http_namespace(Name="mynamespace")["OperationId"]
    namespace_id = client.get_operation(OperationId=operation_id)["Operation"][
        "Targets"
    ]["NAMESPACE"]

    resp = client.create_service(
        Name="my service",
        CreatorRequestId="crid",
        Description="my service",
        DnsConfig={
            "NamespaceId": namespace_id,
            "RoutingPolicy": "WEIGHTED",
            "DnsRecords": [{"Type": "SRV", "TTL": 0}],
        },
        HealthCheckConfig={"Type": "TCP", "ResourcePath": "/sth"},
        HealthCheckCustomConfig={"FailureThreshold": 125},
        Type="HTTP",
    )

    assert "Service" in resp
    assert "Id" in resp["Service"]
    assert "Arn" in resp["Service"]
    assert resp["Service"]["Name"] == "my service"
    assert "NamespaceId" not in resp["Service"]
    assert resp["Service"]["Description"] == "my service"
    assert resp["Service"]["DnsConfig"] == {
        "NamespaceId": namespace_id,
        "RoutingPolicy": "WEIGHTED",
        "DnsRecords": [{"Type": "SRV", "TTL": 0}],
    }
    assert resp["Service"]["HealthCheckConfig"] == {
        "Type": "TCP",
        "ResourcePath": "/sth",
    }
    assert resp["Service"]["HealthCheckCustomConfig"] == {"FailureThreshold": 125}
    assert resp["Service"]["Type"] == "HTTP"
    assert resp["Service"]["CreatorRequestId"] == "crid"


@mock_aws
def test_get_service():
    client = boto3.client("servicediscovery", region_name="ap-southeast-1")

    operation_id = client.create_http_namespace(Name="mynamespace")["OperationId"]
    namespace_id = client.get_operation(OperationId=operation_id)["Operation"][
        "Targets"
    ]["NAMESPACE"]

    service_id = client.create_service(Name="my service", NamespaceId=namespace_id)[
        "Service"
    ]["Id"]

    resp = client.get_service(Id=service_id)

    assert "Service" in resp
    assert "Id" in resp["Service"]
    assert "Arn" in resp["Service"]
    assert resp["Service"]["Name"] == "my service"
    assert resp["Service"]["NamespaceId"] == namespace_id


@mock_aws
def test_get_unknown_service():
    client = boto3.client("servicediscovery", region_name="ap-southeast-1")

    with pytest.raises(ClientError) as exc:
        client.get_service(Id="unknown")
    err = exc.value.response["Error"]
    assert err["Code"] == "ServiceNotFound"
    assert err["Message"] == "unknown"


@mock_aws
def test_delete_service():
    client = boto3.client("servicediscovery", region_name="eu-west-1")

    operation_id = client.create_http_namespace(Name="mynamespace")["OperationId"]
    namespace_id = client.get_operation(OperationId=operation_id)["Operation"][
        "Targets"
    ]["NAMESPACE"]
    service_id = client.create_service(Name="my service", NamespaceId=namespace_id)[
        "Service"
    ]["Id"]

    client.delete_service(Id=service_id)

    with pytest.raises(ClientError) as exc:
        client.get_service(Id=service_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "ServiceNotFound"
    assert err["Message"] == service_id


@mock_aws
def test_update_service_description():
    client = boto3.client("servicediscovery", region_name="ap-southeast-1")
    operation_id = client.create_http_namespace(Name="mynamespace")["OperationId"]
    namespace_id = client.get_operation(OperationId=operation_id)["Operation"][
        "Targets"
    ]["NAMESPACE"]

    service_id = client.create_service(
        Name="my service",
        NamespaceId=namespace_id,
        Description="first desc",
        DnsConfig={
            "NamespaceId": namespace_id,
            "RoutingPolicy": "WEIGHTED",
            "DnsRecords": [{"Type": "SRV", "TTL": 0}],
        },
        HealthCheckConfig={"Type": "TCP", "ResourcePath": "/sth"},
    )["Service"]["Id"]

    client.update_service(Id=service_id, Service={"Description": "updated desc"})

    resp = client.get_service(Id=service_id)

    assert "Service" in resp
    assert resp["Service"]["Id"] == service_id
    assert "Arn" in resp["Service"]
    assert resp["Service"]["Name"] == "my service"
    assert resp["Service"]["NamespaceId"] == namespace_id
    assert resp["Service"]["Description"] == "updated desc"
    # From the docs:
    #    If you omit any existing DnsRecords or HealthCheckConfig
    #    configurations from an UpdateService request,
    #    the configurations are deleted from the service.
    assert "DnsConfig" not in resp["Service"]
    assert resp["Service"]["HealthCheckConfig"] == {
        "Type": "TCP",
        "ResourcePath": "/sth",
    }


@mock_aws
def test_update_service_others():
    client = boto3.client("servicediscovery", region_name="ap-southeast-1")
    operation_id = client.create_http_namespace(Name="mynamespace")["OperationId"]
    namespace_id = client.get_operation(OperationId=operation_id)["Operation"][
        "Targets"
    ]["NAMESPACE"]

    service_id = client.create_service(
        Name="my service",
        NamespaceId=namespace_id,
        Description="first desc",
        DnsConfig={
            "RoutingPolicy": "WEIGHTED",
            "DnsRecords": [{"Type": "SRV", "TTL": 0}],
        },
    )["Service"]["Id"]

    client.update_service(
        Id=service_id,
        Service={
            "DnsConfig": {"DnsRecords": [{"Type": "SRV", "TTL": 12}]},
            "HealthCheckConfig": {"Type": "TCP", "ResourcePath": "/sth"},
        },
    )

    resp = client.get_service(Id=service_id)

    assert "Service" in resp
    assert resp["Service"]["Id"] == service_id
    assert "Arn" in resp["Service"]
    assert resp["Service"]["Name"] == "my service"
    assert resp["Service"]["NamespaceId"] == namespace_id
    assert resp["Service"]["Description"] == "first desc"
    assert resp["Service"]["DnsConfig"] == (
        {"RoutingPolicy": "WEIGHTED", "DnsRecords": [{"Type": "SRV", "TTL": 12}]}
    )
    assert resp["Service"]["HealthCheckConfig"] == {
        "Type": "TCP",
        "ResourcePath": "/sth",
    }



import json
from time import sleep
from unittest import SkipTest
from uuid import uuid4

import boto3
import pytest

from moto import settings

from . import (
    allow_aws_request,
    aws_verified,
    sfn_allow_dynamodb,
    sfn_role_policy,
    verify_execution_result,
)


@aws_verified
@pytest.mark.aws_verified
def test_state_machine_with_simple_failure_state():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Don't need to test this in ServerMode")

    def _verify_result(client, execution, execution_arn):
        assert "stopDate" in execution
        assert execution["error"] == "DefaultStateError"
        assert execution["cause"] == "No Matches!"
        assert execution["input"] == "{}"

        history = client.get_execution_history(executionArn=execution_arn)
        assert len(history["events"]) == 3
        assert [e["type"] for e in history["events"]] == [
            "ExecutionStarted",
            "FailStateEntered",
            "ExecutionFailed",
        ]

        state_machine_arn = execution["stateMachineArn"]
        executions = client.list_executions(stateMachineArn=state_machine_arn)[
            "executions"
        ]
        assert len(executions) == 1
        assert executions[0]["executionArn"] == execution_arn
        assert executions[0]["stateMachineArn"] == state_machine_arn
        assert executions[0]["status"] == "FAILED"

    verify_execution_result(_verify_result, "FAILED", "failure")


@aws_verified
@pytest.mark.aws_verified
def test_state_machine_with_input():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Don't need to test this in ServerMode")

    _input = {"my": "data"}

    def _verify_result(client, execution, execution_arn):
        assert execution["input"] == json.dumps(_input)

    verify_execution_result(
        _verify_result,
        "FAILED",
        "failure",
        exec_input=json.dumps(_input),
    )


@aws_verified
@pytest.mark.aws_verified
def test_state_machine_with_json_regex():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Don't need to test this in ServerMode")

    _input = {"Parameters": [{"Name": "/eggs/cooked/like", "Value": "poached"}]}
    _output = {
        "Parameters": [{"Name": "/eggs/cooked/like", "Value": "poached"}],
        "parameterTransforms": {"eggs": ["poached"]},
    }

    def _verify_result(client, execution, execution_arn):
        assert execution["input"] == json.dumps(_input)
        assert json.loads(execution["output"]) == _output
        return True

    verify_execution_result(
        _verify_result,
        "SUCCEEDED",
        "json_regex",
        exec_input=json.dumps(_input),
    )


@aws_verified
@pytest.mark.aws_verified
def test_version_is_only_available_when_published():
    from tests.test_stepfunctions.test_stepfunctions import simple_definition

    iam = boto3.client("iam", region_name="us-east-1")
    role_name = f"sfn_role_{str(uuid4())[0:6]}"
    sfn_role = iam.create_role(
        RoleName=role_name,
        AssumeRolePolicyDocument=json.dumps(sfn_role_policy),
        Path="/",
    )["Role"]["Arn"]
    iam.put_role_policy(
        PolicyDocument=json.dumps(sfn_allow_dynamodb),
        PolicyName="allowLambdaInvoke",
        RoleName=role_name,
    )
    sleep(10 if allow_aws_request() else 0)

    client = boto3.client("stepfunctions", region_name="us-east-1")

    try:
        name1 = f"sfn_name_{str(uuid4())[0:6]}"
        response = client.create_state_machine(
            name=name1, definition=simple_definition, roleArn=sfn_role
        )
        assert "stateMachineVersionArn" not in response
        arn1 = response["stateMachineArn"]

        resp = client.update_state_machine(
            stateMachineArn=arn1, publish=True, tracingConfiguration={"enabled": True}
        )
        assert resp["stateMachineVersionArn"] == f"{arn1}:1"

        resp = client.update_state_machine(
            stateMachineArn=arn1, tracingConfiguration={"enabled": False}
        )
        assert "stateMachineVersionArn" not in resp

        name2 = f"sfn_name_{str(uuid4())[0:6]}"
        response = client.create_state_machine(
            name=name2, definition=simple_definition, roleArn=sfn_role, publish=True
        )
        arn2 = response["stateMachineArn"]
        assert response["stateMachineVersionArn"] == f"{arn2}:1"

        resp = client.update_state_machine(
            stateMachineArn=arn2, publish=True, tracingConfiguration={"enabled": True}
        )
        assert resp["stateMachineVersionArn"] == f"{arn2}:2"
    finally:
        client.delete_state_machine(stateMachineArn=arn1)
        client.delete_state_machine(stateMachineArn=arn2)
        iam.delete_role_policy(RoleName=role_name, PolicyName="allowLambdaInvoke")
        iam.delete_role(RoleName=role_name)



from typing import Optional, TypedDict

from moto.stepfunctions.parser.api import (
    ActivityFailedEventDetails,
    ActivityScheduledEventDetails,
    ActivityScheduleFailedEventDetails,
    ActivityStartedEventDetails,
    ActivitySucceededEventDetails,
    ActivityTimedOutEventDetails,
    ExecutionAbortedEventDetails,
    ExecutionFailedEventDetails,
    ExecutionStartedEventDetails,
    ExecutionSucceededEventDetails,
    ExecutionTimedOutEventDetails,
    LambdaFunctionFailedEventDetails,
    LambdaFunctionScheduledEventDetails,
    LambdaFunctionScheduleFailedEventDetails,
    LambdaFunctionStartFailedEventDetails,
    LambdaFunctionSucceededEventDetails,
    LambdaFunctionTimedOutEventDetails,
    MapIterationEventDetails,
    MapRunFailedEventDetails,
    MapRunStartedEventDetails,
    MapStateStartedEventDetails,
    StateEnteredEventDetails,
    StateExitedEventDetails,
    TaskFailedEventDetails,
    TaskScheduledEventDetails,
    TaskStartedEventDetails,
    TaskStartFailedEventDetails,
    TaskSubmitFailedEventDetails,
    TaskSubmittedEventDetails,
    TaskSucceededEventDetails,
    TaskTimedOutEventDetails,
)


class EventDetails(TypedDict):
    activityFailedEventDetails: Optional[ActivityFailedEventDetails]
    activityScheduleFailedEventDetails: Optional[ActivityScheduleFailedEventDetails]
    activityScheduledEventDetails: Optional[ActivityScheduledEventDetails]
    activityStartedEventDetails: Optional[ActivityStartedEventDetails]
    activitySucceededEventDetails: Optional[ActivitySucceededEventDetails]
    activityTimedOutEventDetails: Optional[ActivityTimedOutEventDetails]
    taskFailedEventDetails: Optional[TaskFailedEventDetails]
    taskScheduledEventDetails: Optional[TaskScheduledEventDetails]
    taskStartFailedEventDetails: Optional[TaskStartFailedEventDetails]
    taskStartedEventDetails: Optional[TaskStartedEventDetails]
    taskSubmitFailedEventDetails: Optional[TaskSubmitFailedEventDetails]
    taskSubmittedEventDetails: Optional[TaskSubmittedEventDetails]
    taskSucceededEventDetails: Optional[TaskSucceededEventDetails]
    taskTimedOutEventDetails: Optional[TaskTimedOutEventDetails]
    executionFailedEventDetails: Optional[ExecutionFailedEventDetails]
    executionStartedEventDetails: Optional[ExecutionStartedEventDetails]
    executionSucceededEventDetails: Optional[ExecutionSucceededEventDetails]
    executionAbortedEventDetails: Optional[ExecutionAbortedEventDetails]
    executionTimedOutEventDetails: Optional[ExecutionTimedOutEventDetails]
    mapStateStartedEventDetails: Optional[MapStateStartedEventDetails]
    mapIterationStartedEventDetails: Optional[MapIterationEventDetails]
    mapIterationSucceededEventDetails: Optional[MapIterationEventDetails]
    mapIterationFailedEventDetails: Optional[MapIterationEventDetails]
    mapIterationAbortedEventDetails: Optional[MapIterationEventDetails]
    lambdaFunctionFailedEventDetails: Optional[LambdaFunctionFailedEventDetails]
    lambdaFunctionScheduleFailedEventDetails: Optional[
        LambdaFunctionScheduleFailedEventDetails
    ]
    lambdaFunctionScheduledEventDetails: Optional[LambdaFunctionScheduledEventDetails]
    lambdaFunctionStartFailedEventDetails: Optional[
        LambdaFunctionStartFailedEventDetails
    ]
    lambdaFunctionSucceededEventDetails: Optional[LambdaFunctionSucceededEventDetails]
    lambdaFunctionTimedOutEventDetails: Optional[LambdaFunctionTimedOutEventDetails]
    stateEnteredEventDetails: Optional[StateEnteredEventDetails]
    stateExitedEventDetails: Optional[StateExitedEventDetails]
    mapRunStartedEventDetails: Optional[MapRunStartedEventDetails]
    mapRunFailedEventDetails: Optional[MapRunFailedEventDetails]



from typing import Any, Dict, List, Optional, Set, Tuple

from moto.core import DEFAULT_ACCOUNT_ID
from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.config import default_user_config
from moto.core.model_instances import reset_model_data


class MotoAPIBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.proxy_urls_to_passthrough: Set[str] = set()
        self.proxy_hosts_to_passthrough: Set[str] = set()

    def reset(self) -> None:
        region_name = self.region_name
        account_id = self.account_id

        BackendDict.reset()
        reset_model_data()
        self.__init__(region_name, account_id)  # type: ignore[misc]

    def get_transition(self, model_name: str) -> Dict[str, Any]:
        from moto.moto_api import state_manager

        return state_manager.get_transition(model_name)

    def set_transition(self, model_name: str, transition: Dict[str, Any]) -> None:
        from moto.moto_api import state_manager

        state_manager.set_transition(model_name, transition)

    def unset_transition(self, model_name: str) -> None:
        from moto.moto_api import state_manager

        state_manager.unset_transition(model_name)

    def set_athena_result(
        self,
        rows: List[Dict[str, Any]],
        column_info: List[Dict[str, str]],
        account_id: str,
        region: str,
    ) -> None:
        from moto.athena.models import QueryResults, athena_backends

        backend = athena_backends[account_id][region]
        results = QueryResults(rows=rows, column_info=column_info)
        backend.query_results_queue.append(results)

    def set_ce_cost_usage(self, result: Dict[str, Any], account_id: str) -> None:
        from moto.ce.models import ce_backends

        backend = ce_backends[account_id]["global"]
        backend.cost_usage_results_queue.append(result)

    def set_lambda_simple_result(
        self, result: str, account_id: str, region: str
    ) -> None:
        from moto.awslambda_simple.models import lambda_simple_backends

        backend = lambda_simple_backends[account_id][region]
        backend.lambda_simple_results_queue.append(result)

    def set_resilience_result(
        self, result: List[Dict[str, Any]], account_id: str, region: str
    ) -> None:
        from moto.resiliencehub.models import resiliencehub_backends

        backend = resiliencehub_backends[account_id][region]
        backend.app_assessments_queue.append(result)

    def set_sagemaker_result(
        self,
        body: str,
        content_type: str,
        prod_variant: str,
        custom_attrs: str,
        account_id: str,
        region: str,
    ) -> None:
        from moto.sagemakerruntime.models import sagemakerruntime_backends

        backend = sagemakerruntime_backends[account_id][region]
        backend.results_queue.append((body, content_type, prod_variant, custom_attrs))

    def set_sagemaker_async_result(
        self,
        is_failure: bool,
        data: str,
        account_id: str,
        region: str,
    ) -> None:
        from moto.sagemakerruntime.models import sagemakerruntime_backends

        backend = sagemakerruntime_backends[account_id][region]
        backend.async_results_queue.append((is_failure, data))

    def set_rds_data_result(
        self,
        records: Optional[List[List[Dict[str, Any]]]],
        column_metadata: Optional[List[Dict[str, Any]]],
        nr_of_records_updated: Optional[int],
        generated_fields: Optional[List[Dict[str, Any]]],
        formatted_records: Optional[str],
        account_id: str,
        region: str,
    ) -> None:
        from moto.rdsdata.models import QueryResults, rdsdata_backends

        backend = rdsdata_backends[account_id][region]
        backend.results_queue.append(
            QueryResults(
                records=records,
                column_metadata=column_metadata,
                number_of_records_updated=nr_of_records_updated,
                generated_fields=generated_fields,
                formatted_records=formatted_records,
            )
        )

    def set_inspector2_findings_result(
        self,
        results: Optional[List[List[Dict[str, Any]]]],
        account_id: str,
        region: str,
    ) -> None:
        from moto.inspector2.models import inspector2_backends

        backend = inspector2_backends[account_id][region]
        backend.findings_queue.append(results)

    def set_timestream_result(
        self,
        query: Optional[str],
        query_results: List[Dict[str, Any]],
        account_id: str,
        region: str,
    ) -> None:
        from moto.timestreamquery.models import (
            TimestreamQueryBackend,
            timestreamquery_backends,
        )

        backend: TimestreamQueryBackend = timestreamquery_backends[account_id][region]
        if query not in backend.query_result_queue:
            backend.query_result_queue[query] = []
        backend.query_result_queue[query].extend(query_results)

    def get_proxy_passthrough(self) -> Tuple[Set[str], Set[str]]:
        return self.proxy_urls_to_passthrough, self.proxy_hosts_to_passthrough

    def set_proxy_passthrough(
        self, http_urls: List[str], https_hosts: List[str]
    ) -> None:
        for url in http_urls:
            self.proxy_urls_to_passthrough.add(url)
        for host in https_hosts:
            self.proxy_hosts_to_passthrough.add(host)

    def delete_proxy_passthroughs(self) -> None:
        self.proxy_urls_to_passthrough.clear()
        self.proxy_hosts_to_passthrough.clear()

    def get_config(self) -> Dict[str, Any]:
        return {
            "batch": default_user_config["batch"],
            "lambda": default_user_config["lambda"],
        }

    def set_config(self, config: Dict[str, Any]) -> None:
        if "batch" in config:
            default_user_config["batch"] = config["batch"]
        if "lambda" in config:
            default_user_config["lambda"] = config["lambda"]
        if "stepfunctions" in config:
            default_user_config["stepfunctions"] = config["stepfunctions"]


moto_api_backend = MotoAPIBackend(region_name="global", account_id=DEFAULT_ACCOUNT_ID)



from typing import Any, Dict, List

from moto.utilities.utils import get_partition


def parameter_arn(account_id: str, region: str, parameter_name: str) -> str:
    if parameter_name[0] == "/":
        parameter_name = parameter_name[1:]
    return f"arn:{get_partition(region)}:ssm:{region}:{account_id}:parameter/{parameter_name}"


def convert_to_tree(parameters: List[Dict[str, Any]]) -> Dict[str, Any]:
    """
    Convert input into a smaller, less redundant data set in tree form
    Input: [{"Name": "/a/b/c", "Value": "af-south-1", ...}, ..]
    Output: {"a": {"b": {"c": {"Value": af-south-1}, ..}, ..}, ..}
    """
    tree_dict: Dict[str, Any] = {}
    for p in parameters:
        current_level = tree_dict
        for path in p["Name"].split("/"):
            if path == "":
                continue
            if path not in current_level:
                current_level[path] = {}
            current_level = current_level[path]
        current_level["Value"] = p["Value"]
    return tree_dict


def convert_to_params(tree: Dict[str, Any]) -> List[Dict[str, Any]]:
    """
    Inverse of 'convert_to_tree'
    """

    def m(
        tree: Dict[str, Any], params: List[Dict[str, Any]], current_path: str = ""
    ) -> None:
        for key, value in tree.items():
            if key == "Value":
                params.append({"Name": current_path, "Value": value})
            else:
                m(value, params, current_path + "/" + key)

    params: List[Dict[str, Any]] = []
    m(tree, params)
    return params



import itertools
from typing import Any, Dict, List, Optional, Tuple

from moto.core.common_models import CloudFormationModel

from ..exceptions import (
    FlowLogAlreadyExists,
    InvalidAggregationIntervalParameterError,
    InvalidDependantParameterError,
    InvalidDependantParameterTypeError,
    InvalidFlowLogIdError,
)
from ..utils import (
    generic_filter,
    random_flow_log_id,
    utc_date_and_time,
)
from .core import TaggedEC2Resource


class FlowLogs(TaggedEC2Resource, CloudFormationModel):
    def __init__(
        self,
        ec2_backend: Any,
        flow_log_id: str,
        resource_id: str,
        traffic_type: str,
        log_destination: str,
        log_group_name: str,
        deliver_logs_permission_arn: str,
        max_aggregation_interval: str,
        log_destination_type: str,
        log_format: str,
        deliver_logs_status: str = "SUCCESS",
        deliver_logs_error_message: Optional[str] = None,
    ):
        self.ec2_backend = ec2_backend
        self.id = flow_log_id
        self.resource_id = resource_id
        self.traffic_type = traffic_type
        self.log_destination = log_destination
        self.log_group_name = log_group_name
        self.deliver_logs_permission_arn = deliver_logs_permission_arn
        self.deliver_logs_status = deliver_logs_status
        self.deliver_logs_error_message = deliver_logs_error_message
        self.max_aggregation_interval = max_aggregation_interval
        self.log_destination_type = log_destination_type
        self.log_format = log_format

        self.created_at = utc_date_and_time()

    @staticmethod
    def cloudformation_name_type() -> str:
        return ""

    @staticmethod
    def cloudformation_type() -> str:
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-flowlog.html
        return "AWS::EC2::FlowLog"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "FlowLogs":
        from ..models import ec2_backends

        properties = cloudformation_json["Properties"]

        resource_type = properties.get("ResourceType")
        resource_id = [properties.get("ResourceId")]
        traffic_type = properties.get("TrafficType")
        deliver_logs_permission_arn = properties.get("DeliverLogsPermissionArn")
        log_destination_type = properties.get("LogDestinationType")
        log_destination = properties.get("LogDestination")
        log_group_name = properties.get("LogGroupName")
        log_format = properties.get("LogFormat")
        max_aggregation_interval = properties.get("MaxAggregationInterval")

        ec2_backend = ec2_backends[account_id][region_name]
        flow_log, _ = ec2_backend.create_flow_logs(
            resource_type,
            resource_id,
            traffic_type,
            deliver_logs_permission_arn,
            log_destination_type,
            log_destination,
            log_group_name,
            log_format,
            max_aggregation_interval,
        )
        for tag in properties.get("Tags", []):
            tag_key = tag["Key"]
            tag_value = tag["Value"]
            flow_log[0].add_tag(tag_key, tag_value)

        return flow_log[0]

    @property
    def physical_resource_id(self) -> str:
        return self.id

    def get_filter_value(
        self, filter_name: str, method_name: Optional[str] = None
    ) -> Any:
        """
        API Version 2016-11-15 defines the following filters for DescribeFlowLogs:

        * deliver-log-status
        * log-destination-type
        * flow-log-id
        * log-group-name
        * resource-id
        * traffic-type
        * tag:key=value
        * tag-key

        Taken from: https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_DescribeFlowLogs.html
        """
        if filter_name == "resource-id":
            return self.resource_id
        elif filter_name == "traffic-type":
            return self.traffic_type
        elif filter_name == "log-destination-type":
            return self.log_destination_type
        elif filter_name == "flow-log-id":
            return self.id
        elif filter_name == "log-group-name":
            return self.log_group_name
        elif filter_name == "deliver-log-status":
            return "SUCCESS"
        else:
            return super().get_filter_value(filter_name, "DescribeFlowLogs")


class FlowLogsBackend:
    def __init__(self) -> None:
        self.flow_logs: Dict[str, FlowLogs] = {}

    def _validate_request(
        self,
        log_group_name: str,
        log_destination: str,
        log_destination_type: str,
        max_aggregation_interval: str,
        deliver_logs_permission_arn: str,
    ) -> None:
        if log_group_name is None and log_destination is None:
            raise InvalidDependantParameterError(
                "LogDestination", "LogGroupName", "not provided"
            )

        if log_destination_type == "s3":
            if log_group_name is not None:
                raise InvalidDependantParameterTypeError(
                    "LogDestination", "cloud-watch-logs", "LogGroupName"
                )
        elif log_destination_type == "cloud-watch-logs":
            if deliver_logs_permission_arn is None:
                raise InvalidDependantParameterError(
                    "DeliverLogsPermissionArn", "LogDestinationType", "cloud-watch-logs"
                )

        if max_aggregation_interval not in ["60", "600"]:
            raise InvalidAggregationIntervalParameterError(
                "Flow Log Max Aggregation Interval"
            )

    def create_flow_logs(
        self,
        resource_type: str,
        resource_ids: List[str],
        traffic_type: str,
        deliver_logs_permission_arn: str,
        log_destination_type: str,
        log_destination: str,
        log_group_name: str,
        log_format: str,
        max_aggregation_interval: str,
    ) -> Tuple[List[FlowLogs], List[Any]]:
        # Guess it's best to put it here due to possible
        # lack of them in the CloudFormation template
        max_aggregation_interval = (
            "600" if max_aggregation_interval is None else max_aggregation_interval
        )
        log_destination_type = (
            "cloud-watch-logs" if log_destination_type is None else log_destination_type
        )
        log_format = (
            "${version} ${account-id} ${interface-id} ${srcaddr} ${dstaddr} ${srcport} ${dstport} ${protocol} ${packets} ${bytes} ${start} ${end} ${action} ${log-status}"
            if log_format is None
            else log_format
        )

        # Validate the requests paremeters
        self._validate_request(
            log_group_name,
            log_destination,
            log_destination_type,
            max_aggregation_interval,
            deliver_logs_permission_arn,
        )

        flow_logs_set = []
        unsuccessful = []

        for resource_id in resource_ids:
            deliver_logs_status = "SUCCESS"
            deliver_logs_error_message = None
            flow_log_id = random_flow_log_id()
            if resource_type == "VPC":
                # Validate VPCs exist
                self.get_vpc(resource_id)  # type: ignore[attr-defined]
            elif resource_type == "Subnet":
                # Validate Subnets exist
                self.get_subnet(resource_id)  # type: ignore[attr-defined]
            elif resource_type == "NetworkInterface":
                # Validate NetworkInterfaces exist
                self.get_network_interface(resource_id)  # type: ignore[attr-defined]

            if log_destination_type == "s3":
                from moto.s3.exceptions import MissingBucket
                from moto.s3.models import s3_backends

                arn = log_destination.split(":", 5)[5]
                try:
                    s3_backends[self.account_id][self.partition].get_bucket(arn)  # type: ignore[attr-defined]
                except MissingBucket:
                    # Instead of creating FlowLog report
                    # the unsuccessful status for the
                    # given resource_id
                    unsuccessful.append(
                        (resource_id, "400", f"LogDestination: {arn} does not exist.")
                    )
                    continue
            elif log_destination_type == "cloud-watch-logs":
                from moto.logs.exceptions import ResourceNotFoundException
                from moto.logs.models import logs_backends

                # API allows to create a FlowLog with a
                # non-existing LogGroup. It however later
                # on reports the FAILED delivery status.
                try:
                    # Need something easy to check the group exists.
                    # The list_tags_log_group seems to do the trick.
                    logs = logs_backends[self.account_id][self.region_name]  # type: ignore[attr-defined]
                    logs.list_tags_log_group(log_group_name)
                except ResourceNotFoundException:
                    deliver_logs_status = "FAILED"
                    deliver_logs_error_message = "Access error"

            all_flow_logs = self.describe_flow_logs()
            if any(
                (fl.resource_id, fl.log_group_name, fl.log_destination)
                == (resource_id, log_group_name, log_destination)
                for fl in all_flow_logs
            ):
                raise FlowLogAlreadyExists()
            flow_logs = FlowLogs(
                self,
                flow_log_id,
                resource_id,
                traffic_type,
                log_destination,
                log_group_name,
                deliver_logs_permission_arn,
                max_aggregation_interval,
                log_destination_type,
                log_format,
                deliver_logs_status,
                deliver_logs_error_message,
            )
            self.flow_logs[flow_log_id] = flow_logs
            flow_logs_set.append(flow_logs)

        return flow_logs_set, unsuccessful

    def describe_flow_logs(
        self, flow_log_ids: Optional[List[str]] = None, filters: Any = None
    ) -> List[FlowLogs]:
        matches = list(itertools.chain([i for i in self.flow_logs.values()]))
        if flow_log_ids:
            matches = [flow_log for flow_log in matches if flow_log.id in flow_log_ids]
        if filters:
            matches = generic_filter(filters, matches)
        return matches

    def delete_flow_logs(self, flow_log_ids: List[str]) -> None:
        non_existing = []
        for flow_log in flow_log_ids:
            if flow_log in self.flow_logs:
                self.flow_logs.pop(flow_log, None)
            else:
                non_existing.append(flow_log)

        if non_existing:
            raise InvalidFlowLogIdError(
                len(flow_log_ids), " ".join(x for x in flow_log_ids)
            )
